from .utils import *
from .config import *

from .restore_nn import Restore_NN
from .restore_nn import Restore_PCAplusNN
from .suppress_warnings import suppress_warnings

dofftlog_alphas = False

cosmopower_derived_params_names = ['100*theta_s',
                                   'sigma8',
                                   'YHe',
                                   'z_reio',
                                   'Neff',
                                   'tau_rec',
                                   'z_rec',
                                   'rs_rec',
                                   'ra_rec',
                                   'tau_star',
                                   'z_star',
                                   'rs_star',
                                   'ra_star',
                                   'rs_drag']

cp_l_max_scalars = 11000 # max multipole of train ing data

cosmo_model_list = [
    'lcdm',
    'mnu',
    'neff',
    'wcdm',
    'ede',
    'mnu-3states',
    'ede-v2'
]

emulator_dict = {}
emulator_dict['lcdm'] = {}
emulator_dict['mnu'] = {}
emulator_dict['neff'] = {}
emulator_dict['wcdm'] = {}
emulator_dict['ede'] = {}
emulator_dict['mnu-3states'] = {}
emulator_dict['ede-v2'] = {}

### note on ncdm:
# N_ncdm : 3
# m_ncdm : 0.02, 0.02, 0.02
# deg_ncdm: 1
# and
# N_ncdm: 1
# deg_ncdm: 3
# m_ncdm : 0.02
# are equivalent but deg_ncdm: 3 is much faster. 


emulator_dict['lcdm']['TT'] = 'TT_v1'
emulator_dict['lcdm']['TE'] = 'TE_v1'
emulator_dict['lcdm']['EE'] = 'EE_v1'
emulator_dict['lcdm']['PP'] = 'PP_v1'
emulator_dict['lcdm']['PKNL'] = 'PKNL_v1'
emulator_dict['lcdm']['PKL'] = 'PKL_v1'
emulator_dict['lcdm']['PKLFFTLOG_ALPHAS_REAL'] = 'PKLFFTLOGALPHAS_creal_v1'
emulator_dict['lcdm']['PKLFFTLOG_ALPHAS_IMAG'] = 'PKLFFTLOGALPHAS_cimag_v1'
emulator_dict['lcdm']['DER'] = 'DER_v1'
emulator_dict['lcdm']['DAZ'] = 'DAZ_v1'
emulator_dict['lcdm']['HZ'] = 'HZ_v1'
emulator_dict['lcdm']['S8Z'] = 'S8Z_v1'
emulator_dict['lcdm']['default'] = {}
emulator_dict['lcdm']['default']['tau_reio'] = 0.054
emulator_dict['lcdm']['default']['H0'] = 67.66
emulator_dict['lcdm']['default']['ln10^{10}A_s'] = 3.047
emulator_dict['lcdm']['default']['omega_b'] = 0.02242
emulator_dict['lcdm']['default']['omega_cdm'] = 0.11933
emulator_dict['lcdm']['default']['n_s'] = 0.9665
emulator_dict['lcdm']['default']['N_ur'] = 2.0328
emulator_dict['lcdm']['default']['N_ncdm'] = 1
emulator_dict['lcdm']['default']['m_ncdm'] = 0.06

emulator_dict['mnu']['TT'] = 'TT_mnu_v1'
emulator_dict['mnu']['TE'] = 'TE_mnu_v1'
emulator_dict['mnu']['EE'] = 'EE_mnu_v1'
emulator_dict['mnu']['PP'] = 'PP_mnu_v1'
emulator_dict['mnu']['PKNL'] = 'PKNL_mnu_v1'
emulator_dict['mnu']['PKL'] = 'PKL_mnu_v1'
emulator_dict['mnu']['DER'] = 'DER_mnu_v1'
emulator_dict['mnu']['DAZ'] = 'DAZ_mnu_v1'
emulator_dict['mnu']['HZ'] = 'HZ_mnu_v1'
emulator_dict['mnu']['S8Z'] = 'S8Z_mnu_v1'
emulator_dict['mnu']['default'] = {}
emulator_dict['mnu']['default']['tau_reio'] = 0.054
emulator_dict['mnu']['default']['H0'] = 67.66
emulator_dict['mnu']['default']['ln10^{10}A_s'] = 3.047
emulator_dict['mnu']['default']['omega_b'] = 0.02242
emulator_dict['mnu']['default']['omega_cdm'] = 0.11933
emulator_dict['mnu']['default']['n_s'] = 0.9665
emulator_dict['mnu']['default']['N_ur'] = 2.0328
emulator_dict['mnu']['default']['N_ncdm'] = 1
emulator_dict['mnu']['default']['m_ncdm'] = 0.06

emulator_dict['neff']['TT'] = 'TT_neff_v1'
emulator_dict['neff']['TE'] = 'TE_neff_v1'
emulator_dict['neff']['EE'] = 'EE_neff_v1'
emulator_dict['neff']['PP'] = 'PP_neff_v1'
emulator_dict['neff']['PKNL'] = 'PKNL_neff_v1'
emulator_dict['neff']['PKL'] = 'PKL_neff_v1'
emulator_dict['neff']['DER'] = 'DER_neff_v1'
emulator_dict['neff']['DAZ'] = 'DAZ_neff_v1'
emulator_dict['neff']['HZ'] = 'HZ_neff_v1'
emulator_dict['neff']['S8Z'] = 'S8Z_neff_v1'
emulator_dict['neff']['default'] = {}
emulator_dict['neff']['default']['tau_reio'] = 0.054
emulator_dict['neff']['default']['H0'] = 67.66
emulator_dict['neff']['default']['ln10^{10}A_s'] = 3.047
emulator_dict['neff']['default']['omega_b'] = 0.02242
emulator_dict['neff']['default']['omega_cdm'] = 0.11933
emulator_dict['neff']['default']['n_s'] = 0.9665
emulator_dict['neff']['default']['N_ur'] = 2.0328 # this is the default value in class v2 to get Neff = 3.046
emulator_dict['neff']['default']['N_ncdm'] = 1
emulator_dict['neff']['default']['m_ncdm'] = 0.06


emulator_dict['wcdm']['TT'] = 'TT_w_v1'
emulator_dict['wcdm']['TE'] = 'TE_w_v1'
emulator_dict['wcdm']['EE'] = 'EE_w_v1'
emulator_dict['wcdm']['PP'] = 'PP_w_v1'
emulator_dict['wcdm']['PKNL'] = 'PKNL_w_v1'
emulator_dict['wcdm']['PKL'] = 'PKL_w_v1'
emulator_dict['wcdm']['DER'] = 'DER_w_v1'
emulator_dict['wcdm']['DAZ'] = 'DAZ_w_v1'
emulator_dict['wcdm']['HZ'] = 'HZ_w_v1'
emulator_dict['wcdm']['S8Z'] = 'S8Z_w_v1'
emulator_dict['wcdm']['default'] = {}
emulator_dict['wcdm']['default']['tau_reio'] = 0.054
emulator_dict['wcdm']['default']['H0'] = 67.66
emulator_dict['wcdm']['default']['ln10^{10}A_s'] = 3.047
emulator_dict['wcdm']['default']['omega_b'] = 0.02242
emulator_dict['wcdm']['default']['omega_cdm'] = 0.11933
emulator_dict['wcdm']['default']['n_s'] = 0.9665
emulator_dict['wcdm']['default']['N_ur'] = 2.0328 # this is the default value in class v2 to get Neff = 3.046
emulator_dict['wcdm']['default']['N_ncdm'] = 1
emulator_dict['wcdm']['default']['m_ncdm'] = 0.06

emulator_dict['ede']['TT'] = 'TT_v1'
emulator_dict['ede']['TE'] = 'TE_v1'
emulator_dict['ede']['EE'] = 'EE_v1'
emulator_dict['ede']['PP'] = 'PP_v1'
emulator_dict['ede']['PKNL'] = 'PKNL_v1'
emulator_dict['ede']['PKL'] = 'PKL_v1'
emulator_dict['ede']['DER'] = 'DER_v1'
emulator_dict['ede']['DAZ'] = 'DAZ_v1'
emulator_dict['ede']['HZ'] = 'HZ_v1'
emulator_dict['ede']['S8Z'] = 'S8Z_v1'
emulator_dict['ede']['default'] = {}
emulator_dict['ede']['default']['fEDE'] = 0.001
emulator_dict['ede']['default']['tau_reio'] = 0.054
emulator_dict['ede']['default']['H0'] = 67.66
emulator_dict['ede']['default']['ln10^{10}A_s'] = 3.047
emulator_dict['ede']['default']['omega_b'] = 0.02242
emulator_dict['ede']['default']['omega_cdm'] = 0.11933
emulator_dict['ede']['default']['n_s'] = 0.9665
emulator_dict['ede']['default']['log10z_c'] = 3.562 # e.g. from https://github.com/mwt5345/class_ede/blob/master/class/notebooks-ede/2-CMB-Comparison.ipynb
emulator_dict['ede']['default']['thetai_scf'] = 2.83 # e.g. from https://github.com/mwt5345/class_ede/blob/master/class/notebooks-ede/2-CMB-Comparison.ipynb
emulator_dict['ede']['default']['r'] = 0.
emulator_dict['ede']['default']['N_ur'] = 0.00641 # this is the default value in class v2 to get Neff = 3.046
emulator_dict['ede']['default']['N_ncdm'] = 3
emulator_dict['ede']['default']['m_ncdm'] = 0.02


emulator_dict['mnu-3states']['TT'] = 'TT_v1'
emulator_dict['mnu-3states']['TE'] = 'TE_v1'
emulator_dict['mnu-3states']['EE'] = 'EE_v1'
emulator_dict['mnu-3states']['PP'] = 'PP_v1'
emulator_dict['mnu-3states']['PKNL'] = 'PKNL_v1'
emulator_dict['mnu-3states']['PKL'] = 'PKL_v1'
emulator_dict['mnu-3states']['DER'] = 'DER_v1'
emulator_dict['mnu-3states']['DAZ'] = 'DAZ_v1'
emulator_dict['mnu-3states']['HZ'] = 'HZ_v1'
emulator_dict['mnu-3states']['S8Z'] = 'S8Z_v1'
emulator_dict['mnu-3states']['default'] = {}
emulator_dict['mnu-3states']['default']['tau_reio'] = 0.054
emulator_dict['mnu-3states']['default']['H0'] = 67.66
emulator_dict['mnu-3states']['default']['ln10^{10}A_s'] = 3.047
emulator_dict['mnu-3states']['default']['omega_b'] = 0.02242
emulator_dict['mnu-3states']['default']['omega_cdm'] = 0.11933
emulator_dict['mnu-3states']['default']['n_s'] = 0.9665
emulator_dict['mnu-3states']['default']['N_ur'] = 0.00641 # this is the default value in class v2 to get Neff = 3.046
emulator_dict['mnu-3states']['default']['N_ncdm'] = 3
emulator_dict['mnu-3states']['default']['m_ncdm'] = 0.02

emulator_dict['ede-v2']['TT'] = 'TT_v2'
emulator_dict['ede-v2']['TE'] = 'TE_v2'
emulator_dict['ede-v2']['EE'] = 'EE_v2'
emulator_dict['ede-v2']['PP'] = 'PP_v2'
emulator_dict['ede-v2']['PKNL'] = 'PKNL_v2'
emulator_dict['ede-v2']['PKL'] = 'PKL_v2'
emulator_dict['ede-v2']['DER'] = 'DER_v2'
emulator_dict['ede-v2']['DAZ'] = 'DAZ_v2'
emulator_dict['ede-v2']['HZ'] = 'HZ_v2'
emulator_dict['ede-v2']['S8Z'] = 'S8Z_v2'

emulator_dict['ede-v2']['default'] = {}
emulator_dict['ede-v2']['default']['fEDE'] = 0.001
emulator_dict['ede-v2']['default']['tau_reio'] = 0.054
emulator_dict['ede-v2']['default']['H0'] = 67.66
emulator_dict['ede-v2']['default']['ln10^{10}A_s'] = 3.047
emulator_dict['ede-v2']['default']['omega_b'] = 0.02242
emulator_dict['ede-v2']['default']['omega_cdm'] = 0.11933
emulator_dict['ede-v2']['default']['n_s'] = 0.9665
emulator_dict['ede-v2']['default']['log10z_c'] = 3.562 # e.g. from https://github.com/mwt5345/class_ede/blob/master/class/notebooks-ede/2-CMB-Comparison.ipynb
emulator_dict['ede-v2']['default']['thetai_scf'] = 2.83 # e.g. from https://github.com/mwt5345/class_ede/blob/master/class/notebooks-ede/2-CMB-Comparison.ipynb
emulator_dict['ede-v2']['default']['r'] = 0.
emulator_dict['ede-v2']['default']['N_ur'] = 0.00441 # this is the default value in class v3 to get Neff = 3.044
emulator_dict['ede-v2']['default']['N_ncdm'] = 3
emulator_dict['ede-v2']['default']['m_ncdm'] = 0.02


cp_tt_nn = {}
cp_te_nn = {}
cp_ee_nn = {}
cp_pp_nn = {}
cp_pknl_nn = {}
cp_pkl_nn = {}
cp_pkl_fftlog_alphas_real_nn = {}
cp_pkl_fftlog_alphas_imag_nn = {}
cp_pkl_fftlog_alphas_nus = {}
cp_der_nn = {}
cp_da_nn = {}
cp_h_nn = {}
cp_s8_nn = {}

import warnings
from contextlib import contextmanager
import logging

# Suppress absl warnings
import absl.logging
absl.logging.set_verbosity('error')
# Suppress TensorFlow warnings
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
with suppress_warnings():
    import tensorflow as tf
    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)


import re

def split_emulator_string(input_string):
    match = re.match(r"(.+)-v(\d+)", input_string)
    if match:
        folder = match.group(1)
        version = match.group(2)
        return folder, version
    else:
        folder = input_string
        version = '1'
        return folder, version




for mp in cosmo_model_list:
    folder, version = split_emulator_string(mp)
    # print(folder, version)
    path_to_emulators = path_to_class_sz_data + '/' + folder +'/'
    
    cp_tt_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'TTTEEE/' + emulator_dict[mp]['TT'])
    
    cp_te_nn[mp] = Restore_PCAplusNN(restore_filename=path_to_emulators + 'TTTEEE/' + emulator_dict[mp]['TE'])
    
    with suppress_warnings():
        cp_ee_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'TTTEEE/' + emulator_dict[mp]['EE'])
    
    cp_pp_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'PP/' + emulator_dict[mp]['PP'])
    
    cp_pknl_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'PK/' + emulator_dict[mp]['PKNL'])
    
    cp_pkl_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'PK/' + emulator_dict[mp]['PKL'])

    if (mp == 'lcdm') and (dofftlog_alphas == True):
        cp_pkl_fftlog_alphas_real_nn[mp] = Restore_PCAplusNN(restore_filename=path_to_emulators + 'PK/' + emulator_dict[mp]['PKLFFTLOG_ALPHAS_REAL']
                                 )
        cp_pkl_fftlog_alphas_imag_nn[mp] = Restore_PCAplusNN(restore_filename=path_to_emulators + 'PK/' + emulator_dict[mp]['PKLFFTLOG_ALPHAS_IMAG']
                                 )
        cp_pkl_fftlog_alphas_nus[mp] = np.load(path_to_emulators + 'PK/PKL_FFTLog_alphas_nu_v1.npz')
    
    cp_der_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'derived-parameters/' + emulator_dict[mp]['DER'])
    
    cp_da_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'growth-and-distances/' + emulator_dict[mp]['DAZ'])
    
    cp_h_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'growth-and-distances/' + emulator_dict[mp]['HZ'])
    
    cp_s8_nn[mp] = Restore_NN(restore_filename=path_to_emulators + 'growth-and-distances/' + emulator_dict[mp]['S8Z'])
    






from classy_szfast.classy_szfast import Class_szfast
from .config import *
from .utils import *
from .cosmopower import *
from .pks_and_sigmas import *

from pathlib import Path
import sys


from .custom_profiles.custom_profiles import *


from .custom_bias.custom_bias import *



from .utils import *
from .config import *


from .suppress_warnings import suppress_warnings

import warnings
from contextlib import contextmanager
import logging

# Suppress absl warnings
import absl.logging
absl.logging.set_verbosity('error')
# Suppress TensorFlow warnings
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
with suppress_warnings():
    import tensorflow as tf
    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
dtype = tf.float32


class Restore_NN(tf.keras.Model):

    def __init__(self, 
                 parameters=None, 
                 modes=None, 
                 parameters_mean=None, 
                 parameters_std=None, 
                 features_mean=None, 
                 features_std=None, 
                 n_hidden=[512,512,512], 
                 restore=False, 
                 restore_filename=None, 
                 trainable=True,
                 optimizer=None,
                 verbose=False, 
                 ):
        
        # super
        super(Restore_NN, self).__init__()

        # restore
        
        self.restore(restore_filename)


        # input parameters mean and std
        self.parameters_mean = tf.constant(self.parameters_mean_, dtype=dtype, name='parameters_mean')
        self.parameters_std = tf.constant(self.parameters_std_, dtype=dtype, name='parameters_std')

        # (log)-spectra mean and std
        self.features_mean = tf.constant(self.features_mean_, dtype=dtype, name='features_mean')
        self.features_std = tf.constant(self.features_std_, dtype=dtype, name='features_std')

        # weights, biases and activation function parameters for each layer of the network
        self.W = []
        self.b = []
        self.alphas = []
        self.betas = [] 
        for i in range(self.n_layers):
            self.W.append(tf.Variable(tf.random.normal([self.architecture[i], self.architecture[i+1]], 0., 1e-3), name="W_" + str(i), trainable=trainable))
            self.b.append(tf.Variable(tf.zeros([self.architecture[i+1]]), name = "b_" + str(i), trainable=trainable))
        for i in range(self.n_layers-1):
            self.alphas.append(tf.Variable(tf.random.normal([self.architecture[i+1]]), name = "alphas_" + str(i), trainable=trainable))
            self.betas.append(tf.Variable(tf.random.normal([self.architecture[i+1]]), name = "betas_" + str(i), trainable=trainable))

        # restore weights if restore = True

        for i in range(self.n_layers):
            self.W[i].assign(self.W_[i])
            self.b[i].assign(self.b_[i])
        for i in range(self.n_layers-1):
            self.alphas[i].assign(self.alphas_[i])
            self.betas[i].assign(self.betas_[i])

        # optimizer
        self.optimizer = optimizer or tf.keras.optimizers.Adam()
        self.verbose= verbose

        # print initialization info, if verbose
        if self.verbose:
            multiline_str = "\nInitialized cosmopower_NN model, \n" \
                            f"mapping {self.n_parameters} input parameters to {self.n_modes} output modes, \n" \
                            f"using {len(self.n_hidden)} hidden layers, \n" \
                            f"with {list(self.n_hidden)} nodes, respectively. \n"
            print(multiline_str)



    # from https://github.com/HTJense/cosmopower/blob/packaging-paper/cosmopower/cosmopower_NN.py
    def restore(self, filename: str, allow_pickle: bool = False) -> None:
        r"""
        Load pre-trained model.
        The default file format is compressed numpy files (.npz). The
        Module will attempt to use this as a file extension and restore
        from there (i.e. look for `filename.npz`). If this file does
        not exist, and `allow_pickle` is set to True, then the file
        `filename.pkl` will be attempted to be read by `restore_pickle`.

        The function will trim the file extension from `filename`, so
        `restore("filename")` and `restore("filename.npz")` are identical.

        Parameters:
        :param filename: filename (without suffix) where model was saved.
        :param allow_pickle: whether or not to permit passing this filename
                             to the `restore_pickle` function.
        """
        # Check if npz file exists.
        filename_npz = filename + ".npz"
        if not os.path.exists(filename_npz):
            # Can we load this file as a pickle file?
            filename_pkl = filename + ".pkl"
            if allow_pickle and os.path.exists(filename_pkl):
                self.restore_pickle(filename_pkl)
                return

            raise IOError(f"Failed to restore network from {filename}: "
                          + (" is a pickle file, try setting 'allow_pickle = \
                              True'" if os.path.exists(filename_pkl) else
                             " does not exist."))

        with open(filename_npz, "rb") as fp:
            fpz = np.load(fp, allow_pickle=True)["arr_0"].flatten()[0]

            # print('filename_npz:', filename_npz)
            # print("Keys in the .npz file:", list(fpz.keys()))

            self.architecture = fpz["architecture"]
            self.n_layers = fpz["n_layers"]
            self.n_hidden = fpz["n_hidden"]
            self.n_parameters = fpz["n_parameters"]
            self.n_modes = fpz["n_modes"]

            self.parameters = list(fpz["parameters"])
            self.modes = fpz["modes"]

            # self.parameters_mean_ = fpz["parameters_mean"]
            # self.parameters_std_ = fpz["parameters_std"]
            # self.features_mean_ = fpz["features_mean"]
            # self.features_std_ = fpz["features_std"]

            # Attempt to load 'parameters_mean' or fall back to 'param_train_mean'
            self.parameters_mean_ = fpz.get("parameters_mean", fpz.get("param_train_mean"))
            self.parameters_std_ = fpz.get("parameters_std", fpz.get("param_train_std"))
            self.features_mean_ = fpz.get("features_mean", fpz.get("feature_train_mean"))
            self.features_std_ = fpz.get("features_std", fpz.get("feature_train_std"))


            # Fallback to 'weights_' if individual 'W_i' are not found
            if "weights_" in fpz:
                # Assign the list of weight arrays from 'weights_' directly
                self.W_ = fpz["weights_"]
            else:
                # Use individual weight arrays if available
                self.W_ = [fpz[f"W_{i}"] for i in range(self.n_layers)]

            # Fallback to 'biases_' if individual 'b_i' are not found
            if "biases_" in fpz:
                self.b_ = fpz["biases_"]
            else:
                self.b_ = [fpz[f"b_{i}"] for i in range(self.n_layers)]

            self.alphas_ = fpz[f"alphas_"]
            self.betas_ = fpz[f"betas_"]


    # auxiliary function to sort input parameters
    def dict_to_ordered_arr_np(self, 
                               input_dict, 
                               ):
        r"""
        Sort input parameters

        Parameters:
            input_dict (dict [numpy.ndarray]):
                input dict of (arrays of) parameters to be sorted

        Returns:
            numpy.ndarray:
                parameters sorted according to desired order
        """
        if self.parameters is not None:
            return np.stack([input_dict[k] for k in self.parameters], axis=1)
        else:
            return np.stack([input_dict[k] for k in input_dict], axis=1)


    # forward prediction given input parameters implemented in Numpy
    def forward_pass_np(self, 
                        parameters_arr
                        ):
        r"""
        Forward pass through the network to predict the output, 
        fully implemented in Numpy

        Parameters:
            parameters_arr (numpy.ndarray):
                array of input parameters

        Returns:
          numpy.ndarray:
            output predictions
        """
        # forward pass through the network
        act = []
        layers = [(parameters_arr - self.parameters_mean_)/self.parameters_std_]
        for i in range(self.n_layers-1):

            # linear network operation
            act.append(np.dot(layers[-1], self.W_[i]) + self.b_[i])

            # pass through activation function
            layers.append((self.betas_[i] + (1.-self.betas_[i])*1./(1.+np.exp(-self.alphas_[i]*act[-1])))*act[-1])

        # final (linear) layer -> (standardised) predictions
        layers.append(np.dot(layers[-1], self.W_[-1]) + self.b_[-1])

        # rescale and output
        return layers[-1]*self.features_std_ + self.features_mean_


    # Numpy array predictions
    def predictions_np(self, 
                       parameters_dict
                       ):
        r"""
        Predictions given input parameters collected in a dict.
        Fully implemented in Numpy. Calls ``forward_pass_np``
        after ordering the input parameter dict

        Parameters:
            parameters_dict (dict [numpy.ndarray]):
                dictionary of (arrays of) parameters

        Returns:
            numpy.ndarray:
                output predictions
        """
        parameters_arr = self.dict_to_ordered_arr_np(parameters_dict)
        return self.forward_pass_np(parameters_arr)


    # Numpy array 10.**predictions
    def ten_to_predictions_np(self,
                            parameters_dict
                            ):
        r"""
        10^predictions given input parameters collected in a dict.
        Fully implemented in Numpy. It raises 10 to the output
        from ``forward_pass_np``

        Parameters:
            parameters_dict (dict [numpy.ndarray]):
                dictionary of (arrays of) parameters

        Returns:
            numpy.ndarray:
                10^output predictions
        """
        return 10.**self.predictions_np(parameters_dict)


class Restore_PCAplusNN(tf.keras.Model):

    def __init__(self, 
                 cp_pca=None,
                 n_hidden=[512,512,512], 
                 restore=False, 
                 restore_filename=None, 
                 trainable=True, 
                 optimizer=None,
                 verbose=False,
                 ):
        r"""
        Constructor.
        """
        # super
        super(Restore_PCAplusNN, self).__init__()


        self.restore(restore_filename)

        # input parameters mean and std
        self.parameters_mean = tf.constant(self.parameters_mean_, dtype=dtype, name='parameters_mean')
        self.parameters_std = tf.constant(self.parameters_std_, dtype=dtype, name='parameters_std')

        # PCA mean and std
        self.pca_mean = tf.constant(self.pca_mean_, dtype=dtype, name='pca_mean')
        self.pca_std = tf.constant(self.pca_std_, dtype=dtype, name='pca_std')

        # (log)-spectra mean and std
        self.features_mean = tf.constant(self.features_mean_, dtype=dtype, name='features_mean')
        self.features_std = tf.constant(self.features_std_, dtype=dtype, name='features_std')

        # pca transform matrix
        self.pca_transform_matrix = tf.constant(self.pca_transform_matrix_, dtype=dtype, name='pca_transform_matrix')

        # weights, biases and activation function parameters for each layer of the network
        self.W = []
        self.b = []
        self.alphas = []
        self.betas = [] 
        for i in range(self.n_layers):
            self.W.append(tf.Variable(tf.random.normal([self.architecture[i], self.architecture[i+1]], 0., np.sqrt(2./self.n_parameters)), name="W_" + str(i), trainable=trainable))
            self.b.append(tf.Variable(tf.zeros([self.architecture[i+1]]), name = "b_" + str(i), trainable=trainable))
        for i in range(self.n_layers-1):
            self.alphas.append(tf.Variable(tf.random.normal([self.architecture[i+1]]), name = "alphas_" + str(i), trainable=trainable))
            self.betas.append(tf.Variable(tf.random.normal([self.architecture[i+1]]), name = "betas_" + str(i), trainable=trainable))

        # restore weights if restore = True
        for i in range(self.n_layers):
            self.W[i].assign(self.W_[i])
            self.b[i].assign(self.b_[i])
        for i in range(self.n_layers-1):
            self.alphas[i].assign(self.alphas_[i])
            self.betas[i].assign(self.betas_[i])

        self.optimizer = optimizer or tf.keras.optimizers.Adam()
        self.verbose= verbose

        # print initialization info, if verbose
        if self.verbose:
            multiline_str = "\nInitialized cosmopower_PCAplusNN model, \n" \
                            f"mapping {self.n_parameters} input parameters to {self.n_pcas} PCA components \n" \
                            f"and then inverting the PCA compression to obtain {self.n_modes} modes \n" \
                            f"The model uses {len(self.n_hidden)} hidden layers, \n" \
                            f"with {list(self.n_hidden)} nodes, respectively. \n"
            print(multiline_str)



    # from https://github.com/HTJense/cosmopower/blob/packaging-paper/cosmopower/cosmopower_PCAplusNN.py
    def restore(self, filename: str, allow_pickle: bool = False) -> None:
        r"""
        Load pre-trained model.
        The default file format is compressed numpy files (.npz). The
        Module will attempt to use this as a file extension and restore
        from there (i.e. look for `filename.npz`). If this file does
        not exist, and `allow_pickle` is set to True, then the file
        `filename.pkl` will be attempted to be read by `restore_pickle`.

        The function will trim the file extension from `filename`, so
        `restore("filename")` and `restore("filename.npz")` are identical.

        Parameters:
        :param filename: filename (without suffix) where model was saved.
        :param allow_pickle: whether or not to permit passing this filename to
                             the `restore_pickle` function.
        """
        # Check if npz file exists.
        filename_npz = filename + ".npz"
        if not os.path.exists(filename_npz):
            # Can we load this file as a pickle file?
            filename_pkl = filename + ".pkl"
            if allow_pickle and os.path.exists(filename_pkl):
                self.restore_pickle(filename_pkl)
                return

            raise IOError(f"Failed to restore network from {filename}: "
                          + (" is a pickle file, try setting 'allow_pickle = \
                              True'" if os.path.exists(filename_pkl) else
                             " does not exist."))

        with open(filename_npz, "rb") as fp:
            fpz = np.load(fp, allow_pickle=True)["arr_0"].flatten()[0]

            # print('filename_npz:', filename_npz)
            # print("Keys in the .npz file:", list(fpz.keys()))


            self.architecture = fpz["architecture"]
            self.n_layers = fpz["n_layers"]
            self.n_hidden = fpz["n_hidden"]
            self.n_parameters = fpz["n_parameters"]
            self.n_modes = fpz["n_modes"]

            self.parameters = fpz["parameters"]
            self.modes = fpz["modes"]

            # Attempt to load 'parameters_mean' or fall back to 'param_train_mean'
            self.parameters_mean_ = fpz.get("parameters_mean", fpz.get("param_train_mean"))
            self.parameters_std_ = fpz.get("parameters_std", fpz.get("param_train_std"))
            self.features_mean_ = fpz.get("features_mean", fpz.get("feature_train_mean"))
            self.features_std_ = fpz.get("features_std", fpz.get("feature_train_std"))

            # Handle PCA-related keys
            # self.pca_mean_ = fpz["pca_mean"]
            # self.pca_std_ = fpz["pca_std"]
            # self.n_pcas = fpz["n_pcas"]

            self.pca_mean_ = fpz["pca_mean"]
            self.pca_std_ = fpz["pca_std"]
            self.n_pcas = fpz["n_pcas"]
            self.pca_transform_matrix_ = fpz["pca_transform_matrix"]

            # print('n_pcas:', self.n_pcas)

            # filename  = "/Users/boris/Work/CLASS-SZ/SO-SZ/cosmopower-organization/lcdm/TTTEEE/TE_v1.pkl"
            # f = open(filename, 'rb')
            # self.W_, self.b_, self.alphas_, self.betas_, \
            # self.parameters_mean_, self.parameters_std_, \
            # self.pca_mean_, self.pca_std_, \
            # self.features_mean_, self.features_std_, \
            # self.parameters, self.n_parameters, \
            # self.modes, self.n_modes, \
            # self.n_pcas, self.pca_transform_matrix_, \
            # self.n_hidden, self.n_layers, self.architecture = pickle.load(f)

            # print('self.n_pcas:', self.n_pcas)
            # print('self.pca_transform_matrix_:', self.pca_transform_matrix_)
            # print('PCA mean:', self.pca_mean_)
            # print('PCA std:', self.pca_std_)
            # f.close()
            # import sys
            # sys.exit(0)

            # self.pca_transform_matrix_ = fpz["pca_transform_matrix"]        

            # Fallback to 'weights_' if individual 'W_i' are not found
            if "weights_" in fpz:
                # Assign the list of weight arrays from 'weights_' directly
                self.W_ = fpz["weights_"]
            else:
                # Use individual weight arrays if available
                self.W_ = [fpz[f"W_{i}"] for i in range(self.n_layers)]

            # Fallback to 'biases_' if individual 'b_i' are not found
            if "biases_" in fpz:
                self.b_ = fpz["biases_"]
            else:
                self.b_ = [fpz[f"b_{i}"] for i in range(self.n_layers)]

            # Handle alphas and betas
            self.alphas_ = fpz.get("alphas_", [fpz.get(f"alphas_{i}") for i in range(self.n_layers - 1)])
            self.betas_ = fpz.get("betas_", [fpz.get(f"betas_{i}") for i in range(self.n_layers - 1)])



    # auxiliary function to sort input parameters
    def dict_to_ordered_arr_np(self, 
                               input_dict, 
                               ):
        r"""
        Sort input parameters

        Parameters:
            input_dict (dict [numpy.ndarray]):
                input dict of (arrays of) parameters to be sorted

        Returns:
            numpy.ndarray:
                parameters sorted according to desired order
        """
        if self.parameters is not None:
            return np.stack([input_dict[k] for k in self.parameters], axis=1)
        else:
            return np.stack([input_dict[k] for k in input_dict], axis=1)


    # forward prediction given input parameters implemented in Numpy
    def forward_pass_np(self, 
                        parameters_arr,
                        ):
        r"""
        Forward pass through the network to predict the output, 
        fully implemented in Numpy

        Parameters:
            parameters_arr (numpy.ndarray):
                array of input parameters

        Returns:
          numpy.ndarray:
            output predictions
        """
        # forward pass through the network
        act = []
        layers = [(parameters_arr - self.parameters_mean_)/self.parameters_std_]
        for i in range(self.n_layers-1):

            # linear network operation
            act.append(np.dot(layers[-1], self.W_[i]) + self.b_[i])

            # pass through activation function
            layers.append((self.betas_[i] + (1.-self.betas_[i])*1./(1.+np.exp(-self.alphas_[i]*act[-1])))*act[-1])

        # final (linear) layer -> (normalized) PCA coefficients
        layers.append(np.dot(layers[-1], self.W_[-1]) + self.b_[-1])

        # rescale PCA coefficients, multiply out PCA basis -> normalised (log)-spectrum, shift and re-scale (log)-spectrum -> output (log)-spectrum
        return np.dot(layers[-1]*self.pca_std_ + self.pca_mean_, self.pca_transform_matrix_)*self.features_std_ + self.features_mean_


    def predictions_np(self, 
                       parameters_dict,
                       ):
        r"""
        Predictions given input parameters collected in a dict.
        Fully implemented in Numpy. Calls ``forward_pass_np``
        after ordering the input parameter dict

        Parameters:
            parameters_dict (dict [numpy.ndarray]):
                dictionary of (arrays of) parameters

        Returns:
            numpy.ndarray:
                output predictions
        """
        parameters_arr = self.dict_to_ordered_arr_np(parameters_dict)
        return self.forward_pass_np(parameters_arr)


    # 10.**predictions
    def ten_to_predictions_np(self,
                              parameters_dict,
                              ):
        r"""
        10^predictions given input parameters collected in a dict.
        Fully implemented in Numpy. It raises 10 to the output
        from ``forward_pass_np``

        Parameters:
            parameters_dict (dict [numpy.ndarray]):
                dictionary of (arrays of) parameters

        Returns:
            numpy.ndarray:
                10^output predictions
        """
        return 10.**self.predictions_np(parameters_dict)



import numpy as np
def custom1_ux(x,m,z,classy_sz,*params):
    # m is m_delta and delta is the definition passed for custom1 profile
    # x is r/r_delta

    # delta is collected as:
    delta_def = classy_sz.delta_def_custom1()
    delta = classy_sz.get_delta_from_delta_def_at_z(delta_def,z)

    # to get r_delta:
    r_delta = classy_sz.get_r_delta_of_m_delta_at_z(delta,m,z)

    # to get c_delta:
    c_delta = classy_sz.get_c_delta_at_m_and_z(m,z,delta_def)

    # to get x_out:
    x_out = classy_sz.x_out_custom1()


    # example: nfw profile/cmb lensing
    rs = r_delta/c_delta
    xs = x*r_delta/rs
    prof = xs**-1.*(1.+ xs)**-2.
    norm = m*c_delta**3/(np.log(1.+x_out*c_delta)-x_out*c_delta/(1.+x_out*c_delta))

    return norm*prof

def custom1_W(z,classy_sz,*params):
    # e.g., we want (chi_star - chi)/chi
    chi_star = classy_sz.chi_star()
    chi = classy_sz.get_chi(z)
    H = classy_sz.Hubble(z)
    Rho_crit_0 = classy_sz.Rho_crit_0()

    h = classy_sz.h()
    H0 = classy_sz.Hubble(0)


    # example: cmb lensing
    if z==0.:
        w = 1e-100
    else:
        w = 3./2.*(H0/h)**2/Rho_crit_0*(chi/(1.+z))**-1.*(chi_star-chi)/chi_star # 

    return w



import os

import get_cosmopower_emus
import class_sz_data

def get_cosmopower_path():
    get_cosmopower_emus.set()
    return os.getenv('PATH_TO_CLASS_SZ_DATA')

path_to_class_sz_data = get_cosmopower_path()
class_sz_data.get_data_from_class_sz_repo(path_to_class_sz_data)


import numpy as np
from datetime import datetime
import multiprocessing
import time
import functools
import re
from pkg_resources import resource_filename
import os
from scipy import optimize
from scipy.integrate import quad
from scipy.interpolate import interp1d
import math
from numpy import linalg as LA
import mcfit
from mcfit import P2xi
# import cosmopower
# import classy_sz as csz



from scipy.interpolate import LinearNDInterpolator
from scipy.interpolate import CloughTocher2DInterpolator

kb = 1.38064852e-23 #m2 kg s-2 K-1
clight = 299792458. #m/s
hplanck=6.62607004e-34 #m2 kg / s
firas_T0 = 2.728 #pivot temperature used in the Max Lkl Analysis
firas_T0_bf = 2.725 #best-fitting temperature

Tcmb_uk = 2.7255e6

G_newton = 6.674e-11
rho_crit_over_h2_in_GeV_per_cm3 = 1.0537e-5


nu_21_cm_in_GHz =  1./21.1*clight*1.e2/1.e9
x_21_cm = hplanck*nu_21_cm_in_GHz/kb/firas_T0_bf*1.e9

kappa_c = 2.1419 # 4M_2-3M_c see below eq. 9b of https://arxiv.org/pdf/1506.06582.pdf

beta_mu = 2.1923

G1 = np.pi**2./6
G2 = 2.4041
G3 = np.pi**4/15.
a_rho = G2/G3
alpha_mu = 2.*G1/3./G2 # = 1/beta_mu = π^2/18ζ(3) see eq. 4.15 CUSO lectures.

z_mu_era = 3e5
z_y_era = 5e4
z_reio_min = 6
z_reio_max = 25
z_recombination_min = 800
z_recombination_max = 1500

# Physical constants
# ------------------
# Light speed
class Const:
    c_km_s = 299792.458  # speed of light
    h_J_s = 6.626070040e-34  # Planck's constant
    kB_J_K = 1.38064852e-23  # Boltzmann constant



from cobaya.theories.classy import classy
from copy import deepcopy
from typing import NamedTuple, Sequence, Union, Optional
from cobaya.tools import load_module
import logging
import os
import numpy as np
import time

class classy_sz(classy):

    use_class_sz_fast_mode = 1 # this is passed in the yaml file
    use_class_sz_no_cosmo_mode = 0 # this is passed in the yaml file
    lensing_lkl = 'ACT'

    baos = None 

    def initialize(self):
        """Importing CLASS from the correct path, if given, and if not, globally."""

        self.classy_module = self.is_installed()
        if not self.classy_module:
            raise NotInstalledError(
                self.log, "Could not find CLASS_SZ. Check error message above.")
        
        from classy_sz import Class, CosmoSevereError, CosmoComputationError
        
        global CosmoComputationError, CosmoSevereError

        self.classy = Class()
        super(classy,self).initialize()


        # Add general CLASS stuff
        self.extra_args["output"] = self.extra_args.get("output", "")

        if "sBBN file" in self.extra_args:
            self.extra_args["sBBN file"] = (
                self.extra_args["sBBN file"].format(classy=self.path))
            
        # Derived parameters that may not have been requested, but will be necessary later
        self.derived_extra = []
        self.log.info("Initialized!")


        ## rename some parameters to avoid conflices
        classy_sz_renames = {

            'omega_m':'Omega_m',
            'Omegam':'Omega_m',
            'Omega_m':'Omega_m'
        }
        self.renames.update(classy_sz_renames)


        if self.use_class_sz_no_cosmo_mode == 1:

            self.log.info("Initializing cosmology part!")

            initial_parameters = self.extra_args.copy()


            self.classy.set(initial_parameters)
            self.classy.compute_class_szfast(likelihood_mode=True)
            self.log.info("cosmology part initialized!")




    def must_provide(self, **requirements):



        if "Cl_sz" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_sz")
            # specify the method to collect the new observable
            self.collectors["Cl_sz"] = Collector(
                    method="cl_sz", # name of the method in classy.pyx
                    args_names=[],
                    args=[])

        if "Cl_yxg" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_yxg")
            # specify the method to collect the new observable
            self.collectors["Cl_yxg"] = Collector(
                    method="cl_yg", # name of the method in classy.pyx
                    args_names=[],
                    args=[])

        if "Cl_gxg" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_gxg")
            # specify the method to collect the new observable
            self.collectors["Cl_gxg"] = Collector(
                    method="cl_gg", # name of the method in classy.pyx
                    args_names=[],
                    args=[])


        if "Cl_gxmu" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_gxmu")
            # specify the method to collect the new observable
            self.collectors["Cl_gxmu"] = Collector(
                    method="cl_gm", # name of the method in classy.pyx
                    args_names=[],
                    args=[])


        if "Cl_muxmu" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_muxmu")
            # specify the method to collect the new observable
            self.collectors["Cl_muxmu"] = Collector(
                    method="cl_mm", # name of the method in classy.pyx
                    args_names=[],
                    args=[])


        if "Cl_kxg" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_kxg")
            # specify the method to collect the new observable
            self.collectors["Cl_kxg"] = Collector(
                    method="cl_kg", # name of the method in classy.pyx
                    args_names=[],
                    args=[])

        if "Cl_kxmu" in requirements:
        # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_kxmu")
        # specify the method to collect the new observable
            self.collectors["Cl_kxmu"] = Collector(
                    method="cl_km", # name of the method in classy.pyx
                    args_names=[],
                    args=[])


        if "Cl_yxmu" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_yxmu")
            # specify the method to collect the new observable
            self.collectors["Cl_yxmu"] = Collector(
                    method="cl_ym", # name of the method in classy.pyx
                    args_names=[],
                    args=[])
        if "Cl_kgxg" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_kgxg")
            # specify the method to collect the new observable
            self.collectors["Cl_kgxg"] = Collector(
                    method="cl_ggamma", # name of the method in classy.pyx
                    args_names=[],
                    args=[])
        if "Cl_kgxmu" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("Cl_kgxmu")
                # specify the method to collect the new observable
                self.collectors["Cl_kgxmu"] = Collector(
                        method="cl_kg_m", # name of the method in classy.pyx
                        args_names=[],
                        args=[])
        if "Cl_IAxg" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("Cl_IAxg")
            # specify the method to collect the new observable
            self.collectors["Cl_IAxg"] = Collector(
                    method="cl_IA_g", # name of the method in classy.pyx
                    args_names=[],
                    args=[])

        if "sz_binned_cluster_counts" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("sz_binned_cluster_counts")
            # specify the method to collect the new observable
            self.collectors["sz_binned_cluster_counts"] = Collector(
                    method="dndzdy_theoretical", # name of the method in classy.pyx
                    args_names=[],
                    args=[])

        if "sz_unbinned_cluster_counts" in requirements:
            # make sure cobaya still runs as it does for standard classy
            requirements.pop("sz_unbinned_cluster_counts")
            # specify the method to collect the new observable
            self.collectors["sz_unbinned_cluster_counts"] = Collector(
                    method="szcounts_ntot_rates_loglike", # name of the method in classy.pyx
                    args_names=[],
                    args=[])

        if "cl_cib_kappa" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("cl_cib_kappa")
                # specify the method to collect the new observable
                self.collectors["cl_cib_kappa"] = Collector(
                        method="cl_lens_cib", # name of the method in classy.pyx
                        args_names=[],
                        args=[])
        if "cl_galn_galn" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("cl_galn_galn")
                # specify the method to collect the new observable
                self.collectors["cl_galn_galn"] = Collector(
                        method="cl_galn_galn", # name of the method in classy.pyx
                        args_names=[],
                        args=[])
        if "cl_galn_lens" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("cl_galn_lens")
                # specify the method to collect the new observable
                self.collectors["cl_galn_lens"] = Collector(
                        method="cl_galn_lens", # name of the method in classy.pyx
                        args_names=[],
                        args=[])

        if "Cl_galnxtsz" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("Cl_galnxtsz")
                # specify the method to collect the new observable
                self.collectors["Cl_galnxtsz"] = Collector(
                        method="cl_galn_tsz", # name of the method in classy.pyx
                        args_names=[],
                        args=[])
        if "Cl_galnxgallens" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("Cl_galnxgallens")
                # specify the method to collect the new observable
                self.collectors["Cl_galnxgallens"] = Collector(
                        method="cl_galn_gallens", # name of the method in classy.pyx
                        args_names=[],
                        args=[])
        if "Cl_lensmagnxtsz" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("Cl_lensmagnxtsz")
                # specify the method to collect the new observable
                self.collectors["Cl_lensmagnxtsz"] = Collector(
                        method="cl_lensmagn_tsz", # name of the method in classy.pyx
                        args_names=[],
                        args=[])
        if "Cl_lensmagnxgallens" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("Cl_lensmagnxgallens")
                # specify the method to collect the new observable
                self.collectors["Cl_lensmagnxgallens"] = Collector(
                        method="cl_lensmagn_gallens", # name of the method in classy.pyx
                        args_names=[],
                        args=[])
        if "Cl_galnxIA" in requirements:
                # make sure cobaya still runs as it does for standard classy
                requirements.pop("Cl_galnxIA")
                # specify the method to collect the new observable
                self.collectors["Cl_galnxIA"] = Collector(
                        method="cl_galn_IA", # name of the method in classy.pyx
                        args_names=[],
                        args=[])
        super().must_provide(**requirements)


    # get the required new observable
    def get_Cl(self, ell_factor=False, units="FIRASmuK2"):
        
        if self.use_class_sz_fast_mode:

            return self.get_Clfast(ell_factor=ell_factor)
        
        else:
        
            return self._get_Cl(ell_factor=ell_factor, units=units, lensed=True)

    def get_Clfast(self,ell_factor = False):

        cls = {}
        cls = deepcopy(self._current_state["Cl"])

        lcp = np.asarray(cls['ell'])

        if ell_factor==True:
            cls['tt'] *= (2.7255e6)**2.*(lcp*(lcp+1.))/2./np.pi
            cls['te'] *= (2.7255e6)**2.*(lcp*(lcp+1.))/2./np.pi
            cls['ee'] *= (2.7255e6)**2.*(lcp*(lcp+1.))/2./np.pi

        if self.lensing_lkl == "ACT" and ell_factor == False:
            cls['tt'] *= (2.7255e6)**2.
            cls['te'] *= (2.7255e6)**2.
            cls['ee'] *= (2.7255e6)**2.

        if self.lensing_lkl ==  "SOLikeT":
            cls['pp'] *= (lcp*(lcp+1.))**2./4.

        elif self.lensing_lkl == "ACT":
            cls['pp'] *= 1.
             
        else: # here for the planck lensing lkl, using lfactor option gives:
            cls['pp'] *= (lcp*(lcp+1.))**2.*1./2./np.pi

        return cls
    
    # get the required new observable
    def get_Cl_sz(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_sz"])
        return cls

    def get_Cl_yxg(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_yxg"])
        return cls
    def get_Cl_kxg(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_kxg"])
        return cls
    def get_Cl_gxg(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_gxg"])
        return cls
    def get_Cl_muxmu(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_muxmu"])
        return cls
    def get_Cl_gxmu(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_gxmu"])
        return cls
    def get_Cl_kxmu(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_kxmu"])
        return cls
    def get_Cl_yxmu(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_yxmu"])
        return cls
    def get_Cl_kgxg(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_kgxg"])
        return cls
    def get_Cl_kgxmu(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_kgxmu"])
        return cls
    def get_Cl_IAxg(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_IAxg"])
        return cls
    def get_cl_galn_lens(self):
        cls = {}
        cls = deepcopy(self._current_state["cl_galn_lens"])
        return cls
    def get_cl_galn_galn(self):
        cls = {}
        cls = deepcopy(self._current_state["cl_galn_galn"])
        return cls
    def get_Cl_galnxtsz(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_galnxtsz"])
        return cls
    def get_Cl_galnxgallens(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_galnxgallens"])
        return cls
    def get_Cl_lensmagnxtsz(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_lensmagnxtsz"])
        return cls
    def get_Cl_lensmagnxgallens(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_lensmagnxgallens"])
        return cls
    def get_Cl_galnxIA(self):
        cls = {}
        cls = deepcopy(self._current_state["Cl_galnxIA"])
        return cls

    # get the required new observable
    def get_sz_unbinned_cluster_counts(self):

        cls = deepcopy(self._current_state["sz_unbinned_cluster_counts"])

        return cls['loglike'],cls['ntot'],cls['rates']


    # get the required new observable
    def get_sz_binned_cluster_counts(self):
        cls = {}
        cls = deepcopy(self._current_state["sz_binned_cluster_counts"])
        return cls

    def get_cl_cib_kappa(self):
        cls = {}
        cls = deepcopy(self._current_state["cl_cib_kappa"])
        return cls

    # IMPORTANT: this method is imported from cobaya and modified to accomodate the emulators
    def calculate(self, state, want_derived=True, **params_values_dict):

        
        params_values = params_values_dict.copy()
        # if baos are requested we need to update the relevant flags
        if self.baos:
            params_values.update({'skip_chi':0,'skip_hubble':0})

        if 'N_ncdm' in self.extra_args.keys():

            if self.extra_args['N_ncdm'] ==  3:
            
                str_mncdm = str(params_values['m_ncdm'])
                params_values['m_ncdm'] = str_mncdm+','+str_mncdm+','+str_mncdm

        try:
            
            params_values['ln10^{10}A_s'] = params_values.pop("logA")
            
            self.set(params_values)
        
        except KeyError:
        
            self.set(params_values)

        # Compute!
        try:
            if self.use_class_sz_fast_mode == 1:


                if self.use_class_sz_no_cosmo_mode == 1:

                    start_time = time.time()
                    self.classy.compute_class_sz(params_values)
                    end_time = time.time()
                    # self.log.info("Execution time of class_sz: {:.5f} seconds".format(end_time - start_time))

                else:

                    start_time = time.time()
                    self.classy.compute_class_szfast(likelihood_mode=True)
                    end_time = time.time()
                    # self.log.info("Execution time of class_szfast: {:.5f} seconds".format(end_time - start_time))
                    # print('pars in classy',self.classy.pars)
            else:

                self.classy.compute()

        # "Valid" failure of CLASS: parameters too extreme -> log and report
        except self.classy_module.CosmoComputationError as e:
            if self.stop_at_error:
                self.log.error(
                    "Computation error (see traceback below)! "
                    "Parameters sent to CLASS: %r and %r.\n"
                    "To ignore this kind of error, make 'stop_at_error: False'.",
                    state["params"], dict(self.extra_args))
                raise
            else:
                self.log.debug("Computation of cosmological products failed. "
                               "Assigning 0 likelihood and going on. "
                               "The output of the CLASS error was %s" % e)
            return False
        # CLASS not correctly initialized, or input parameters not correct
        except self.classy_module.CosmoSevereError:
            self.log.error("Serious error setting parameters or computing results. "
                           "The parameters passed were %r and %r. To see the original "
                           "CLASS' error traceback, make 'debug: True'.",
                           state["params"], self.extra_args)
            raise  # No LoggedError, so that CLASS traceback gets printed
        # Gather products
        for product, collector in self.collectors.items():
            # print(product,collector)
            # Special case: sigma8 needs H0, which cannot be known beforehand:
            if "sigma8" in self.collectors:
                self.collectors["sigma8"].args[0] = 8 / self.classy.h()
            method = getattr(self.classy, collector.method)
            arg_array = self.collectors[product].arg_array
            if isinstance(arg_array, int):
                arg_array = np.atleast_1d(arg_array)
            if arg_array is None:
                state[product] = method(
                    *self.collectors[product].args, **self.collectors[product].kwargs)
            elif isinstance(arg_array, Sequence) or isinstance(arg_array, np.ndarray):
                arg_array = np.array(arg_array)
                if len(arg_array.shape) == 1:
                    # if more than one vectorised arg, assume all vectorised in parallel
                    n_values = len(self.collectors[product].args[arg_array[0]])
                    state[product] = np.zeros(n_values)
                    args = deepcopy(list(self.collectors[product].args))
                    for i in range(n_values):
                        for arg_arr_index in arg_array:
                            args[arg_arr_index] = \
                                self.collectors[product].args[arg_arr_index][i]
                        state[product][i] = method(
                            *args, **self.collectors[product].kwargs)
                elif len(arg_array.shape) == 2:
                    if len(arg_array) > 2:
                        raise NotImplementedError("Only 2 array expanded vars so far.")
                    # Create outer combinations
                    x_and_y = np.array(np.meshgrid(
                        self.collectors[product].args[arg_array[0, 0]],
                        self.collectors[product].args[arg_array[1, 0]])).T
                    args = deepcopy(list(self.collectors[product].args))
                    result = np.empty(shape=x_and_y.shape[:2])
                    for i, row in enumerate(x_and_y):
                        for j, column_element in enumerate(x_and_y[i]):
                            args[arg_array[0, 0]] = column_element[0]
                            args[arg_array[1, 0]] = column_element[1]
                            result[i, j] = method(
                                *args, **self.collectors[product].kwargs)
                    state[product] = (
                        self.collectors[product].args[arg_array[0, 0]],
                        self.collectors[product].args[arg_array[1, 0]], result)
                else:
                    raise ValueError("arg_array not correctly formatted.")
            elif arg_array in self.collectors[product].kwargs:
                value = np.atleast_1d(self.collectors[product].kwargs[arg_array])
                state[product] = np.zeros(value.shape)
                for i, v in enumerate(value):
                    kwargs = deepcopy(self.collectors[product].kwargs)
                    kwargs[arg_array] = v
                    state[product][i] = method(
                        *self.collectors[product].args, **kwargs)
            else:
                raise LoggedError(self.log, "Variable over which to do an array call "
                                            f"not known: arg_array={arg_array}")
            if collector.post:
                state[product] = collector.post(*state[product])
        # Prepare derived parameters

        d, d_extra = self._get_derived_all(derived_requested=want_derived)

        if want_derived:

            state["derived"] = {p: d.get(p) for p in self.output_params}


        state["derived_extra"] = deepcopy(d_extra)



    @classmethod
    def is_installed(cls, **kwargs):
        return load_module('classy_sz')


# this just need to be there as it's used to fill-in self.collectors in must_provide:
class Collector(NamedTuple):
    method: str
    args: Sequence = []
    args_names: Sequence = []
    kwargs: dict = {}
    arg_array: Union[int, Sequence] = None
    post: Optional[callable] = None



import numpy as np

def custom1_b(z,classy_sz,*params):
    # # e.g., we want (chi_star - chi)/chi
    # chi_star = classy_sz.chi_star()
    # chi = classy_sz.get_chi(z)
    # H = classy_sz.Hubble(z)
    # Rho_crit_0 = classy_sz.Rho_crit_0()

    # h = classy_sz.h()
    # H0 = classy_sz.Hubble(0)


    # # example: cmb lensing
    # if z==0.:
    #     w = 1e-100
    # else:
    #     w = 3./2.*(H0/h)**2/Rho_crit_0*(chi/(1.+z))**-1.*(chi_star-chi)/chi_star # 

    return 4.# + 2*z
